/*************************************************************************
*
*	File: pcx.cpp
*	Copyright(c) Andreas Magnusson 2002
*
*	Should work on both big-endian (eg. Sun Sparc) and little-endian (eg. Intel x86)
*	machines.
*
*	DISCLAIMER AND STUFF:
*	This source-code is supplied AS IS. I cannot be held responsible for any damage caused
*	by use or misuse of this source-code nor any derivatives of it.
*	So use the code as you please, but give credit where they are due.
*
*************************************************************************/

// disable the annoying warnings that comes up when you use map...
#pragma warning( disable : 4786)
#include <stdio.h>
#include "image.h"

struct pcx_header
{
  unsigned char manufacturer;
  unsigned char version;
  unsigned char encoding;
  unsigned char bits_per_pixel;
  unsigned short x, y;
  unsigned short width, height;
  unsigned short horz_res;
  unsigned short vert_res;
  unsigned char ega_palette[48];
  unsigned char reserved;
  unsigned char num_color_planes;
  unsigned short bytes_per_line;
  unsigned short palette_type;
  unsigned char padding[58];
	//
	unsigned short make_word(unsigned char *&buf)
	{
		unsigned short rv = *buf++;
		rv += (*buf++ << 8);
		return rv;
	}
	void word2byte(unsigned short v, unsigned char *&buf)
	{
		*buf++ = v & 0xff;
		*buf++ = (v >> 8) & 0xff;
	}
	//
	pcx_header(unsigned char *temphead)
	{
	  // this procedure is needed for compatibility between
		// little-endian machines (eg Intel-based) and big-endian
		// machines (eg Sun Sparc)
	  manufacturer = *temphead++;
		version = *temphead++;
	  encoding = *temphead++;
		bits_per_pixel = *temphead++;
	  x = make_word(temphead);
		y = make_word(temphead);
	  width = make_word(temphead);
		height = make_word(temphead);

	  horz_res = make_word(temphead);
		vert_res = make_word(temphead);

	  // I doubt I will ever use EGA-palettes so I don't copy their
		// contents to the header.
	  //  char ega_palette[48];
		temphead += sizeof(ega_palette);

		reserved = *temphead++;
	  num_color_planes = *temphead++;
		bytes_per_line = make_word(temphead);
		palette_type = make_word(temphead);
		//
	}
	pcx_header()
	{
	}
	void copy_to_buffer(unsigned char *tempbuf)
	{
		//memcpy(tempbuf, this, sizeof(pcx_header));
		memset(tempbuf, 0, sizeof(pcx_header));
		*tempbuf++ = manufacturer;
		*tempbuf++ = version;
		*tempbuf++ = encoding;
		*tempbuf++ = bits_per_pixel;
		word2byte(x, tempbuf);
		word2byte(y, tempbuf);
		word2byte(width, tempbuf);
		word2byte(height, tempbuf);
		word2byte(horz_res, tempbuf);
		word2byte(vert_res, tempbuf);
		tempbuf += sizeof(ega_palette);
		*tempbuf++ = reserved;
		*tempbuf++ = num_color_planes;
		word2byte(bytes_per_line, tempbuf);
		word2byte(palette_type, tempbuf);
	}
};

///////////////////////////////////////////////////////////////////////////////////////

class pcx_image : public image
{
	unsigned char *read_line24(unsigned char *buf, unsigned long line, unsigned long sv);
	unsigned char *write_line24(unsigned char *buf, unsigned long line, unsigned long sv);
public:
	pcx_image();
	virtual bool load(const std::string &file_name);
	virtual bool save(const std::string &file_name);
	virtual const char *get_type() {return "pcx";}
};

// register pcx_image
namespace
{
	register_in_factory<image, pcx_image> register_me("pcx");
};

pcx_image::pcx_image()
{
}

// this function currently only reads 24-bit pcx-files
bool pcx_image::load(const std::string &file_name)
{
	FILE *file;

	file = fopen(file_name.c_str(), "rb");
	if(!file)
		return false;

	fseek(file, 0, SEEK_END);
	long file_size = ftell(file) - sizeof(pcx_header);
	rewind(file);

	unsigned char *temphead = new unsigned char[sizeof(pcx_header)];
	fread(temphead, 1, sizeof(pcx_header), file);
	pcx_header header(temphead);
	delete[] temphead;

	width = header.width + 1;
	height = header.height + 1;
	color_depth = header.num_color_planes << 3;

	long size = header.bytes_per_line * get_height() * header.num_color_planes;
	if(buffer)
		delete[] buffer;

	unsigned char *tmpbuf = new unsigned char[file_size];
	unsigned char *svtmpbuf = tmpbuf; // to delete tmpbuf later...
  
	buffer = new unsigned char[size * 2];
	if(buffer == NULL || tmpbuf == NULL)
	{
		fclose(file);
		return false;
	}

	memset(buffer, 0, size);
	fread(tmpbuf, 1, file_size, file);
	fclose(file);

	// the data is stored as bitplanes in the pcx-image, thus first a row of red, then a row
	// of green and finally a row of blue. Then comes next row of red...theoretically this
	// could be used to save alpha-values in pcx-images, but ASFAIK no paint-program supports
	// that...
	for(int i = 0; i < get_height(); i++)
	{
		for(int j = 0; j < (get_color_depth() >> 3); j++)
			tmpbuf = read_line24(tmpbuf, i * width, j);
	}

	delete[] svtmpbuf;
	return true;
}

bool pcx_image::save(const std::string &file_name)
{
	if(!buffer)
		return false;

	long size = get_width() * get_height() * (get_color_depth() >> 3);
	unsigned char *pack = NULL, *tmppack;

	// hopefully the amount of bytes packed will never
	// be bigger than the original size, but if it does
	// try making it bigger. OBSERVE that it will if your
	// image consists of a lot of vertical lines and the
	// lines are drawn with a color above 191
	// so to be on the safe side we allocate twice that...
	// we should of course process the image and calculate
	// the number of bytes required...
	pack = new unsigned char[size * 2];
	if(!pack)
		return false;
	tmppack = pack;

	FILE *file;

	file = fopen(file_name.c_str(), "wb");
	if(!file)
	{
		delete[] pack;
		return false;
	}

	// setup pcxheader to standard values
  unsigned char *temphead = new unsigned char[sizeof(pcx_header)];
	pcx_header header;
	header.manufacturer = 10;
	header.version = 5;
	header.encoding = 1;
	header.bits_per_pixel = 8;
	header.x = 0;
	header.y = 0;
	header.height = get_height() - 1;
	header.width = get_width() - 1;

	header.horz_res = 71; // ???
	header.vert_res = 71; // ???
	header.bytes_per_line = get_width();
	header.reserved = 0;
	header.num_color_planes = get_color_depth() >> 3;
	header.palette_type = 1;
	header.copy_to_buffer(temphead);

	fwrite(temphead, 1, sizeof(pcx_header), file);

	for(int i = 0; i < get_height(); i++)
	{
		for(int j = 0; j < (get_color_depth() >> 3); j++)
			tmppack = write_line24(tmppack, i * get_width(), j);
	}

	fwrite(pack, 1, (long)tmppack - (long)pack, file);

	fclose(file);
	delete[] temphead;
	delete[] pack;
	return true;
}

// this writes an 24-bit PCX-encoded line used by save() (above)
unsigned char *pcx_image::write_line24(unsigned char *buf, unsigned long line, unsigned long in_pos)
{
	const unsigned long color_planes = get_color_depth() >> 3;
	unsigned char *read_buf = &buffer[line * color_planes];
	unsigned char last = read_buf[in_pos];
	int bpl = 0;
	int read_pos = in_pos + color_planes;
	short runlen = 1;
	for(int i = 1; i < get_width(); i++)
	{
		if(read_buf[read_pos] == last && runlen < 63)
		{
			runlen++;
			if(runlen == 63)
			{
				*buf++ = (unsigned char)((runlen | 0xc0) & 0xff);
				*buf++ = last;
				runlen = 0;
				bpl += 2;
			}
		}
		else
		{
			if(runlen > 1 || 0xc0 == (last & 0xc0))
			{
				*buf++ = (unsigned char)((runlen | 0xc0) & 0xff);
				bpl++;
			}
			*buf++ = last;
			bpl++;
			last = read_buf[read_pos];
			runlen = 1;
		}
		read_pos += color_planes;
	}
	// handle the last pixel too
	if(runlen)
	{
		if(runlen > 1 || 0xc0 == (last & 0xc0))
		{
			*buf++ = (unsigned char)((runlen | 0xc0) & 0xff);
			bpl++;
		}
		*buf++ = last;
		bpl++;
	}
	return buf;
}


unsigned char *pcx_image::read_line24(unsigned char *buf, unsigned long line, unsigned long sv)
{
	const unsigned long color_planes = get_color_depth() >> 3;
	unsigned long byte_count = 0, bytes_read = 0, write_pos = sv;
  unsigned char data;
  unsigned char *buffer24 = (unsigned char *)&buffer[line * color_planes];
  while(bytes_read < get_width())
  {
    data = *buf++;
    if((data >= 192) && (data <= 255))
    {
      byte_count = data - 192;
      data  = *buf++;
      while(byte_count-- > 0)
      {
				buffer24[write_pos] = data;
				write_pos += color_planes;
				bytes_read++;
      }
    }
    else
    {
      buffer24[write_pos] = data;
      write_pos += color_planes;
      bytes_read++;
    }
  }
  return buf;
}
